#include <stdio.h>
#include <math.h>
#include "myrandom.h"

int no_machines;   /* Anzahl der Maschinen im Problem */
int no_jobs;       /* Anzahl der Jobs im Problem */
int **zeiten;      /* Bearbeitungszeiten fuer Job pro Machine */
int **reihenfolge; /* Reihenfolge der Machinen fuer jeden Job */

int **population;  /* Permutationen der Population */
int *fitness;      /* Fitness zu den Permutationen */

int *tmp_jobtime;  /* Feld zur Fitnessberechnung */
int *tmp_machtime; /* Feld zur Fitnessberechnung */
int *tmp_nextmach; /* Feld zur Fitnessberechnung */
int *tmp_rekomb;   /* Feld die Rekombination */

int bestfitness;
int bestindex;

void printusage() 
{
  fprintf(stderr,"hybridEA <options>:\n");
  fprintf(stderr," -p <population size>\n");
  fprintf(stderr," -g <generations>\n");
  fprintf(stderr," -m <mutation rate>\n");
  fprintf(stderr," -g <crossover rate>\n");
  fprintf(stderr," -s <seed>\n");
  fprintf(stderr,"Data file name: jssp.dat\n");
}

void readproblem() 
{
  FILE *fdesc;
  int i,j;

  fdesc = fopen("jssp.dat","r");
  fscanf(fdesc,"%d\n",&no_jobs);
  fscanf(fdesc,"%d\n",&no_machines);
  zeiten = new (int*)[no_jobs];
  reihenfolge = new (int*)[no_jobs];
  for (i=0; i<no_jobs; i++) {
    zeiten[i] = new int[no_machines];
    reihenfolge[i] = new int[no_machines];
  }
  for (i=0; i<no_jobs; i++) {
    for (j=0; j<no_machines; j++) {
      fscanf(fdesc,"%d",&(reihenfolge[i][j]));
      fscanf(fdesc,"%d",&(zeiten[i][j]));
    }
    fscanf(fdesc,"\n");
  }
  fclose(fdesc);

  tmp_jobtime = new (int)[no_jobs];
  tmp_nextmach = new (int)[no_jobs];
  tmp_machtime = new (int)[no_machines];
  tmp_rekomb = new int[no_jobs*no_machines];
}

int computefitness(int *ind)
{
  int i;
  int job,mach,slot,endtime;
  int maxtime = 0;
  
  for (i=0; i<no_jobs; i++) {
    tmp_jobtime[i] = 0;
    tmp_nextmach[i] = 0;
  }
  for (i=0; i<no_machines; i++) {
    tmp_machtime[i] = 0;
  }

  for (i=0; i<no_jobs*no_machines; i++) {
    job = ind[i] / no_machines;
    mach = tmp_nextmach[job]++;
    if (tmp_jobtime[job] > tmp_machtime[reihenfolge[job][mach]])
      slot = tmp_jobtime[job];
    else 
      slot = tmp_machtime[reihenfolge[job][mach]];
    endtime = slot + zeiten[job][mach];
    tmp_jobtime[job] = endtime;
    tmp_machtime[reihenfolge[job][mach]] = endtime;
    if (endtime > maxtime)
      maxtime = endtime;
  }
  return maxtime;
}

void printresult(int *ind)
{
    int i;
  int job,mach,slot,endtime;
  int maxtime = 0;
  
  for (i=0; i<no_jobs; i++) {
    tmp_jobtime[i] = 0;
    tmp_nextmach[i] = 0;
  }
  for (i=0; i<no_machines; i++) {
    tmp_machtime[i] = 0;
  }

  for (i=0; i<no_jobs*no_machines; i++) {
    job = ind[i] / no_machines;
    mach = tmp_nextmach[job]++;
    if (tmp_jobtime[job] > tmp_machtime[reihenfolge[job][mach]])
      slot = tmp_jobtime[job];
    else 
      slot = tmp_machtime[reihenfolge[job][mach]];
    endtime = slot + zeiten[job][mach];
    tmp_jobtime[job] = endtime;
    tmp_machtime[reihenfolge[job][mach]] = endtime;
    if (endtime > maxtime)
      maxtime = endtime;
    printf("Job %d auf Maschine %d - Startzeit: %d, Endzeit: %d\n",
	   job, reihenfolge[job][mach], slot, endtime);
  }
}

void validperm(int *ind)
{
  int i,j;
  int falsch = 0;

  for (i=0; i<no_jobs*no_machines; i++) {
    for (j=i+1; j<no_jobs*no_machines; j++) {
      if (ind[i] == ind[j])
	falsch = 1;
    }
  }
  if (falsch)
    printf("Invalid permutation! \n");
  else 
    printf("permutation ok\n");
}

void initpop(int size)
{
  int i,j,tmp,rnd;

  fitness = new int[size];
  population = new (int*)[size];
  for (i=0; i<size; i++) {
    population[i] = new int[no_jobs * no_machines];
  }

  for (i=0; i<size; i++) {
    for (j=0; j<no_jobs * no_machines; j++) {
      population[i][j] = j;
    }
    for (j=0; j<no_jobs * no_machines; j++) {
      rnd = randint(no_jobs * no_machines);
      if (rnd >= no_jobs * no_machines) {
	printf("Random error\n"); exit;
      }
      tmp = population[i][j];
      population[i][j] = population[i][rnd];
      population[i][rnd] = tmp;
    }
    fitness[i] = computefitness(population[i]);
    if (i==0) {
      bestindex = 0;
      bestfitness = fitness[0];
    } else {
      if (fitness[i] < bestfitness) {
	bestfitness = fitness[i];
	bestindex = i;
      }
    }
  }
}

void rekombination_order(int *first, int *second)
{
  int cut1, cut2, tmp, i, j;

  cut1 = randint(no_jobs * no_machines);
      if (cut1 >= no_jobs * no_machines) {
	printf("Random error\n"); exit;
      }
  cut2 = randint(no_jobs * no_machines);
      if (cut2 >= no_jobs * no_machines) {
	printf("Random error\n"); exit;
      }
  if (cut1>cut2) {
    tmp = cut1;
    cut1 = cut2;
    cut2 = tmp;
  }

  for (i=0; i<no_jobs * no_machines; i++)
    tmp_rekomb[i] = 0;
  for (i=cut1; i<=cut2; i++)
    tmp_rekomb[first[i]] = 1;

  i=0;
  for (j=0; j<cut1; j++) {
    while (tmp_rekomb[second[i]])
      i++;
    first[j] = second[i];
    i++;
  }
  for (j=cut2+1; j<no_jobs * no_machines; j++) {
    while (tmp_rekomb[second[i]])
      i++;
    first[j] = second[i];
    i++;
  }
}

void mutation_switch(int *parent)
{
  int pos1,pos2,tmp,i;

  pos1 = randint(no_jobs * no_machines);
      if (pos1 >= no_jobs * no_machines) {
	printf("Random error\n"); exit;
      }
  pos2 = randint(no_jobs * no_machines);
      if (pos2 >= no_jobs * no_machines) {
	printf("Random error\n"); exit;
      }
  tmp = parent[pos1];
  parent[pos1] = parent[pos2];
  parent[pos2] = tmp;
}



int main (int argc, char* argv[])
{
  int arg_index = 1;
  int tmp_popsize = 15;
  int tmp_generations = 100;
  double tmp_mrate = 0.2;
  double tmp_crate = 0.8;
  long tmp_randomseed = 1;

  int *newind;
  int gen;
  int newfit;
  int worstindex;
  int worstfitness;

  int exitflag = 0;

  int i;
  int index1, index2;

  if (argc > 1) {
    /* parsing */
    while (arg_index < argc) {
      if (strcmp(argv[arg_index], "-p") == 0) {
	arg_index++;
	tmp_popsize = (int) strtol(argv[arg_index++],NULL,10);
      } else {
	if (strcmp(argv[arg_index], "-g") == 0) {
	  arg_index++;
	  tmp_generations = (int) strtol(argv[arg_index++],NULL,10);
	} else {
	  if (strcmp(argv[arg_index], "-s") == 0) {
	    arg_index++;
	    tmp_randomseed = (long) strtol(argv[arg_index++],NULL,10);
	  } else {
	    if (strcmp(argv[arg_index], "-m") == 0) {
	      arg_index++;
	      tmp_mrate = (double) strtod(argv[arg_index++],
					  NULL);
	    } else {
	      if (strcmp(argv[arg_index], "-c") == 0) {
		arg_index++;
		tmp_crate = (double) strtod(argv[arg_index++],NULL);
	      } else {
		arg_index++;
		      fprintf(stderr,"unknown option\n");
		      exitflag = 1;
	      }
	    }
	  }
	}
      }
    }
  }

  if (!exitflag) {

    /* Initialisierung */
    setrandomseed(tmp_randomseed);

    /* Einlesen des Problems */
    readproblem();

    /* Initialisierung der Population */
    initpop(tmp_popsize);

    /* Initialisierung eines lokalen Individuums */
    newind = new int[no_jobs * no_machines];

    /* Evolutionszyklus */
    for (gen=0; gen<tmp_generations; gen++) {
      
      /* Auswahl der Eltern */
      index1 = randint(tmp_popsize);
      index2 = randint(tmp_popsize);

      /* Kopieren des ersten Individuums */
      for (i=0; i<no_jobs * no_machines; i++) {
	newind[i] = population[index1][i];
      }

      /* Rekombination */
      if (randdouble(1.0) < tmp_crate) {
	rekombination_order(newind,population[index2]);
      }

      /* Mutation */
      if (randdouble(1.0) < tmp_mrate) {
	mutation_switch(newind);
      }

      newfit = computefitness(newind);

      /* Bestimme den schlechtesten in der Population */
      worstindex = 0;
      worstfitness = fitness[0];
      for (i=1;i<tmp_popsize; i++) {
	if (fitness[i] > worstfitness) {
	  worstfitness = fitness[i];
	  worstindex = i;
	}
      }

      /* Ersetze das schlechteste Individuum */
      for (i=0;i<no_jobs * no_machines; i++) {
	population[worstindex][i] = newind[i];
      }
      fitness[worstindex] = newfit;

      /* Update beste Fitness */
      if (newfit < bestfitness) {
	bestfitness = newfit;
	bestindex = worstindex;
	printf("Generation %d: Beste Fitness: %d\n",gen,bestfitness);
      }

    }
    printresult(population[bestindex]);
  } else {
    printusage();
  }

}
