
#include <math.h>
#include "myrandom.h"

#define IA 16807
#define IM 2147483647
#define AM (1.0 / IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

#define IB1 1
#define IB2 2
#define IB5 16
#define IB18 131072
#define MASK (IB1+IB2+IB5)

long myrandom_seed;
unsigned long myrandom_seed2;

float ran1(long *idum) 
{
  int j;
  long k;
  static long iy=0;
  static long iv[NTAB];
  float temp;

  if (*idum <= 0 || !iy) {
    if (-(*idum) < 1)
      *idum = 1;
    else
      *idum = -(*idum);
    for (j = NTAB + 7; j >= 0; j--) {
      k = (*idum)/IQ;
      *idum = IA * (*idum - k * IQ) - IR * k;
      if (*idum < 0)
	*idum += IM;
      if (j < NTAB)
	iv[j] = *idum;
    }
    iy = iv[0];
  }
  k = (*idum) / IQ;
  *idum = IA * (*idum - k * IQ) - IR * k;
  if (*idum < 0) 
    *idum += IM;
  j = iy / NDIV;
  iy = iv[j];
  iv[j] = *idum;
  if ((temp = AM * iy) > RNMX)
    return RNMX;
  else
    return temp;
}

int irbit2(unsigned long *iseed)
{
  if (*iseed & IB18) {
    *iseed = ((*iseed ^ MASK) << 1) | IB1;
    return 1;
  } else {
    *iseed <<= 1;
    return 0;
  }
}

double gasdev(long *idum)
{
  static int iset=0;
  static float gset;
  float fac,rsq,v1,v2;

  if (*idum < 0) iset=0;
  if (iset == 0) {
    do {
      v1 = 2.0*ran1(idum)-1.0;
      v2 = 2.0*ran1(idum)-1.0;
      rsq = v1*v1 + v2*v2;
    } while (rsq >= 1.0 || rsq == 0.0);
    fac = sqrt(-2.0*log(rsq)/rsq);
    gset = v1*fac;
    iset = 1;
    return v2*fac;
  } else {
    iset = 0;
    return gset;
  }
}

void setrandomseed(long s)
{
  float tmp;
  if (s > 0) {
    myrandom_seed = -s;
    myrandom_seed2 = s;
  } else {
    if (s == 0) {
      myrandom_seed = -1;
      myrandom_seed2 = 1;
    } else {
      myrandom_seed = s;
      myrandom_seed2 = -s;
    }
  }
  tmp = ran1(&myrandom_seed);
}

int randint(int u)
{
  return (int)((u*1.0)*ran1(&myrandom_seed));
}

int randbit()
{
  return irbit2(&myrandom_seed2);
}

double randdouble(double u)
{
  return u * ran1(&myrandom_seed);
}

double randgaussian()
{
  return gasdev(&myrandom_seed);
}
